/* DDRAM map hd44780 compatible LCD

  00 01 02 03 04 05 06 07 08 09 0A 0B 0C 0D 0E 0F 10 11 12 13 14 15 16 17 18 19 1A 1B 1C 1D 1E 1F 20 21 22 23 24 25 26 27
 | <--             16x4 line 1               --> | <--              16x4 line 3              --> |

  40 41 42 43 44 45 46 47 48 49 4A 4B 4C 4D 4E 4F 50 51 52 53 54 55 56 57 58 59 5A 5B 5C 5D 5E 5F 60 61 62 63 64 65 66 67 
 | <--             16x4 line 2               --> | <--              16x4 line 4              --> |

*/

#if !defined(_COMMON_H_)
    #include <htc.h>
    #include "lcd.h"
    #include "delay.h"
#endif

#if defined(LCD_SHIFT_REGISTER)

void lcdInit(void)
{
    delayMs(50);		 // StartUp delay to voltage rise to regular level

	LCD_EN_DIR = 0;
	LCD_DT_DIR = 0;
	LCD_CK_DIR = 0;

	LCD_EN = 0;
	LCD_DT = 0;
	LCD_CK = 0;

	lcdCmd(0x30);         // interface is now 8-bits long
    lcdCmd(0x38);         // 2 line LCD with 5x8 chars
    lcdCmd(0x0C);         // (0x0C - display on, cursor off, blink off, 0x0F - display on, cursor on, blink on)
    lcdCmd(0x06);         // increment mode, no display shift
    lcdCmd(0x01);         // clear display, (and set DDRAM address to 0x00)
}

void lcdPutch(unsigned char c)
{
    static struct {
		unsigned mode  :1;
        unsigned delay :1;
    } lcd = {1,0};

	unsigned char index = 8;

    if(c == LCD_CMDCHAR)
    {
        if(lcd.mode)
        {
            lcd.mode = 0;
            return;
        }
        else
            lcd.mode = 1;
    }

    if((c < 4) && (lcd.mode == 0))
        lcd.delay = 1;

    while(index)
	{
        if(c & 0x80)
            LCD_DT = 1;
        else
            LCD_DT = 0;

        LCD_CK = 1;
        LCD_CK = 0;

        c = c << 1;
        index--;
    }

    if(lcd.mode)
        LCD_DT = 1;     // LCD_DT act here as LCD_RS pin
    else
        LCD_DT = 0;     // LCD_DT act here as LCD_RS pin

	LCD_EN = 1;
    NOP();
    LCD_EN = 0;

    if(lcd.delay)
        {delayMs(2);}
	#if (_XTAL_FREQ > 12000000)
	else
		{delayUs(45);}
	#endif

    lcd.mode = 1;
    lcd.delay = 0;
}

#else  // LCD_4BIT

void lcdInit(void)
{
    delayMs(50);		 // StartUp delay to voltage rise to regular level
                         // take care that analog functions of used pins are off !!!
    LCD_DIR_D7 = 0;
    LCD_DIR_D6 = 0;
    LCD_DIR_D5 = 0;
    LCD_DIR_D4 = 0;
    LCD_DIR_RS = 0;
    LCD_DIR_EN = 0;

    LCD_EN = 0;
    LCD_RS = 0;          // instruction mode

    #warning "move this delay into main function"
/*
    LCD_D7 = 0;
    LCD_D6 = 0;
    LCD_D5 = 1;
    LCD_D4 = 0;
    LCD_EN = 1; NOP(); LCD_EN = 0; // Interface is now 4bit long
                                   // No delay here.
    LCD_EN = 1; NOP(); LCD_EN = 0; // IMPORTANT !!! Repeat one more time command for 4bit mode
    delayUs(40);                   // Standard delay
*/
    LCD_RS = 1;         // LCD_RS have to be high because of "lcd_putch()" function
    lcdCmd(0x22);       // IMPORTANT !!! Effectively send two times command for 4bit mode (0x20), see above
    lcdCmd(0x28);       // 2 line LCD 5x8 chars
    lcdCmd(0x0C);       // Display On, Cursor off, Blik Cursor off
    lcdCmd(0x06);       // Increment mode, no dispaly shift
    lcdCmd(0x01);       // Clear display, (and set DDRAM address to 0x00)
}


void lcdPutch(unsigned char c)
{
    if (c == LCD_CMDCHAR)
    {
        if (LCD_RS)
        {
            LCD_RS = 0;
            return;
        }
        else
            LCD_RS = 1;
    }

    // high nibble first
    LCD_D7 = 0; if(c & 0x80) LCD_D7 = 1;
    LCD_D6 = 0; if(c & 0x40) LCD_D6 = 1; 
    LCD_D5 = 0; if(c & 0x20) LCD_D5 = 1;
    LCD_D4 = 0; if(c & 0x10) LCD_D4 = 1;
    LCD_EN = 1; NOP(); LCD_EN = 0;
    // low nibble then
    LCD_D7 = 0; if(c & 0x08) LCD_D7 = 1;
    LCD_D6 = 0; if(c & 0x04) LCD_D6 = 1; 
    LCD_D5 = 0; if(c & 0x02) LCD_D5 = 1;
    LCD_D4 = 0; if(c & 0x01) LCD_D4 = 1;
    LCD_EN = 1; NOP(); LCD_EN = 0;

	if((LCD_RS == 0) && (c < 0x04)) // If LCD is in Instruction mode and executed instruction is clear display
		delayMs(2);		            // (0x01) or return home (0x02, 0x03) activate required long delay.
	else							
		delayUs(40);				// standard delay otherwise

	LCD_RS = 1;
}

#endif  // #if defined(LCD_SHIFT_REGISTER)


void lcdPuts(unsigned char c, const unsigned char *s)
{
    if(c != 0)
        lcdCmd(c);

    while(*s)
    {
        lcdPutch(*s);
        s++;
    }
}

/*
void lcdWellcomeScreen(void)
{
    const unsigned char * row1 = "  2x16 hd44780  ";
    const unsigned char * row2 = " compatible LCD ";

    lcd_puts(LCD_LINE1, row1);
    lcd_puts(LCD_LINE2, row2);
}
*/
